// ==============================================================
//                 ORBITER MODULE: protontower
//
// protontower.cpp
// Control module for protontower vessel class
//
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"

// ==============================================================
// Some vessel parameters
// ==============================================================

const double TOWER_SPEED = 0.025;
const double MACHOIRES_SPEED = 0.1;
const double FUELMASS = 10000.0;
const double ISP = 1e10;
const double MAXTHRUST = 0;
VECTOR3 prtattpos = _V(0,-5,-8);


class protontower: public VESSEL2 {
public:
	protontower (OBJHANDLE hVessel, int flightmodel)
		: VESSEL2 (hVessel, flightmodel) {}
	void DefineAnimations (void);
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	void SetSmoke(void);

	THRUSTER_HANDLE th_steam1, th_steam2, th_steam3, th_steam4, th_steam5;


private:
	UINT anim_twr, anim_mach, anim_etage1, anim_etage2, anim_etage3, anim_etage4, anim_etage5;
	double twr_proc, mach_proc, etage1_proc, etage2_proc, etage3_proc, etage4_proc, etage5_proc;
	UINT anim_etage6, anim_etage7, anim_etage8, anim_etage9, anim_etage10;
	double etage6_proc, etage7_proc, etage8_proc, etage9_proc, etage10_proc;

	ANIMATIONCOMPONENT_HANDLE parent1;
	ANIMATIONCOMPONENT_HANDLE parentA1,parentA2,parentA3,parentA4,parentA5,parentA6,parentA7,parentA8,parentA9;

	int twr_status, mach_status, etage1_status, etage2_status, etage3_status, etage4_status, etage5_status;
	int etage6_status, etage7_status, etage8_status, etage9_status;

	ATTACHMENTHANDLE prt_attach;
	double stepdt;
};
// Define animation sequences for moving parts
// --------------------------------------------------------------
void protontower::DefineAnimations (void)
{
	// 1. Tower
	anim_twr = CreateAnimation (0);
	static UINT TwrGrp[16] = {0,3,5,36,38,68,1,4,6,35,66,78,2,7,67,37};

	static MGROUP_TRANSLATE  TwrGrpTrans (0, TwrGrp, 16, _V(0, 0, 140));
	parent1 = AddAnimationComponent (anim_twr, 0, 1, &TwrGrpTrans);


	// 2. Etages
	anim_etage1 = CreateAnimation (0);
	static UINT Etage1Grp[3] = {11,42,70};
	static MGROUP_TRANSLATE  Etage1Trans (0, Etage1Grp, 3, _V(0, -11.8, 0));
	parentA1 = AddAnimationComponent (anim_etage1, 0, 1,   &Etage1Trans, parent1);

	anim_etage2 = CreateAnimation (0);
	static UINT Etage2Grp[3] = {14,45,71};
	static MGROUP_TRANSLATE  Etage2Trans (0, Etage2Grp, 3, _V(0, -11.8, 0));
	parentA2 = AddAnimationComponent (anim_etage2, 0, 1,   &Etage2Trans, parent1);

	anim_etage3 = CreateAnimation (0);
	static UINT Etage3Grp[3] = {17,48,72};
	static MGROUP_TRANSLATE  Etage3Trans (0, Etage3Grp, 3, _V(0, -11.8, 0));
	parentA3 = AddAnimationComponent (anim_etage3, 0, 1,   &Etage3Trans, parent1);

	anim_etage4 = CreateAnimation (0);
	static UINT Etage4Grp[3] = {20,51,73};
	static MGROUP_TRANSLATE  Etage4Trans (0, Etage4Grp, 3, _V(0, -11.8, 0));
	parentA4 = AddAnimationComponent (anim_etage4, 0, 1,   &Etage4Trans, parent1);

	anim_etage5 = CreateAnimation (0);
	static UINT Etage5Grp[3] = {23,54,74};
	static MGROUP_TRANSLATE  Etage5Trans (0, Etage5Grp, 3, _V(0, -11.7, 0));
	parentA5 = AddAnimationComponent (anim_etage5, 0, 1,   &Etage5Trans, parent1);

	anim_etage6 = CreateAnimation (0);
	static UINT Etage6Grp[3] = {26,57,75};
	static MGROUP_TRANSLATE  Etage6Trans (0, Etage6Grp, 3, _V(0, -9, 0));
	parentA6 = AddAnimationComponent (anim_etage6, 0, 1,   &Etage6Trans, parent1);

	anim_etage7 = CreateAnimation (0);
	static UINT Etage7Grp[3] = {29,60,76};
	static MGROUP_TRANSLATE  Etage7Trans (0, Etage7Grp, 3, _V(0, -2, 0));
	parentA7 = AddAnimationComponent (anim_etage7, 0, 1,   &Etage7Trans, parent1);

	anim_etage8 = CreateAnimation (0);
	static UINT Etage8Grp[3] = {8,39,69};
	static MGROUP_TRANSLATE  Etage8Trans (0, Etage8Grp, 3, _V(0, -2, 0));
	parentA8 = AddAnimationComponent (anim_etage8, 0, 1,   &Etage8Trans, parent1);

	anim_etage9 = CreateAnimation (0);
	static UINT Etage9Grp[3] = {33,64,77};
	static MGROUP_TRANSLATE  Etage9Trans (0, Etage9Grp, 3, _V(0, -10, 0));
	parentA9 = AddAnimationComponent (anim_etage9, 0, 1,   &Etage9Trans, parent1);



	// 3. Machoires
	anim_mach = CreateAnimation (0);

	static UINT MachGrpA1[2] = {13,44};
	static MGROUP_ROTATE  MachRotA1 (0, MachGrpA1, 2, _V( 5,35,-5.3), _V(0,-1,0), 0.698);
	AddAnimationComponent (anim_mach, 0, 1,   &MachRotA1, parentA1);

	static UINT MachGrpA2[2] = {16,47};
	static MGROUP_ROTATE  MachRotA2 (0, MachGrpA2, 2, _V( 5,35,-5.3), _V(0,-1,0), 0.698);
	AddAnimationComponent (anim_mach, 0, 1,   &MachRotA2, parentA2);

	static UINT MachGrpA3[2] = {19,50};
	static MGROUP_ROTATE  MachRotA3 (0, MachGrpA3, 2, _V( 5,35,-5.3), _V(0,-1,0), 0.698);
	AddAnimationComponent (anim_mach, 0, 1,   &MachRotA3, parentA3);

	static UINT MachGrpA4[2] = {22,53};
	static MGROUP_ROTATE  MachRotA4 (0, MachGrpA4, 2, _V( 5,35,-5.3), _V(0,-1,0), 0.698);
	AddAnimationComponent (anim_mach, 0, 1,   &MachRotA4, parentA4);

	static UINT MachGrpA5[2] = {25,56};
	static MGROUP_ROTATE  MachRotA5 (0, MachGrpA5, 2, _V( 5,35,-5.3), _V(0,-1,0), 0.698);
	AddAnimationComponent (anim_mach, 0, 1,   &MachRotA5, parentA5);

	static UINT MachGrpA6[2] = {28,59};
	static MGROUP_ROTATE  MachRotA6 (0, MachGrpA6, 2, _V( 5,35,-5.3), _V(0,-1,0), 0.698);
	AddAnimationComponent (anim_mach, 0, 1,   &MachRotA6, parentA6);

	static UINT MachGrpA7[2] = {31,62};
	static MGROUP_ROTATE  MachRotA7 (0, MachGrpA7, 2, _V( 5,35,-5.3), _V(0,-1,0), 0.698);
	AddAnimationComponent (anim_mach, 0, 1,   &MachRotA7, parentA7);

	static UINT MachGrpA8[2] = {10,41};
	static MGROUP_ROTATE  MachRotA8 (0, MachGrpA8, 2, _V( 5,35,-5.3), _V(0,-1,0), 0.698);
	AddAnimationComponent (anim_mach, 0, 1,   &MachRotA8, parentA8);

	static UINT MachGrpA9[2] = {34,65};
	static MGROUP_ROTATE  MachRotA9 (0, MachGrpA9, 2, _V( 5,35,-5.3), _V(0,-1,0), 0.698);
	AddAnimationComponent (anim_mach, 0, 1,   &MachRotA9, parentA9);

	///////// B
	static UINT MachGrpB1[2] = {12,43};
	static MGROUP_ROTATE  MachRotB1 (0, MachGrpB1, 2, _V( -5,35,-5.3), _V(0,1,0), 0.698);
	AddAnimationComponent (anim_mach, 0, 1,   &MachRotB1, parentA1);

	static UINT MachGrpB2[2] = {15,46};
	static MGROUP_ROTATE  MachRotB2 (0, MachGrpB2, 2, _V( -5,35,-5.3), _V(0,1,0), 0.698);
	AddAnimationComponent (anim_mach, 0, 1,   &MachRotB2, parentA2);

	static UINT MachGrpB3[2] = {18,49};
	static MGROUP_ROTATE  MachRotB3 (0, MachGrpB3, 2, _V( -5,35,-5.3), _V(0,1,0), 0.698);
	AddAnimationComponent (anim_mach, 0, 1,   &MachRotB3, parentA3);

	static UINT MachGrpB4[2] = {21,52};
	static MGROUP_ROTATE  MachRotB4 (0, MachGrpB4, 2, _V( -5,35,-5.3), _V(0,1,0), 0.698);
	AddAnimationComponent (anim_mach, 0, 1,   &MachRotB4, parentA4);

	static UINT MachGrpB5[2] = {24,55};
	static MGROUP_ROTATE  MachRotB5 (0, MachGrpB5, 2, _V( -5,35,-5.3), _V(0,1,0), 0.698);
	AddAnimationComponent (anim_mach, 0, 1,   &MachRotB5, parentA5);

	static UINT MachGrpB6[2] = {27,58};
	static MGROUP_ROTATE  MachRotB6 (0, MachGrpB6, 2, _V( -5,35,-5.3), _V(0,1,0), 0.698);
	AddAnimationComponent (anim_mach, 0, 1,   &MachRotB6, parentA6);

	static UINT MachGrpB7[2] = {30,61};
	static MGROUP_ROTATE  MachRotB7 (0, MachGrpB7, 2, _V( -5,35,-5.3), _V(0,1,0), 0.698);
	AddAnimationComponent (anim_mach, 0, 1,   &MachRotB7, parentA7);

	static UINT MachGrpB8[2] = {9,40};
	static MGROUP_ROTATE  MachRotB8 (0, MachGrpB8, 2, _V( -5,35,-5.3), _V(0,1,0), 0.698);
	AddAnimationComponent (anim_mach, 0, 1,   &MachRotB8, parentA8);

	static UINT MachGrpB9[2] = {32,63};
	static MGROUP_ROTATE  MachRotB9 (0, MachGrpB9, 2, _V( -5,35,-5.3), _V(0,1,0), 0.698);
	AddAnimationComponent (anim_mach, 0, 1,   &MachRotB9, parentA9);



}


// ==============================================================
// protontower class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void protontower::clbkSetClassCaps (FILEHANDLE cfg)
{

	stepdt = 0.0;
	twr_proc = 0.0;
	twr_status = 0;
	mach_proc = 0.0;
	mach_status = 0;
	etage1_proc = 0.0;
	etage1_status = 0;
	etage2_proc = 0.0;
	etage2_status = 0;
	etage3_proc = 0.0;
	etage3_status = 0;
	etage4_proc = 0.0;
	etage4_status = 0;
	etage5_proc = 0.0;
	etage5_status = 0;
	etage6_proc = 0.0;
	etage6_status = 0;
	etage7_proc = 0.0;
	etage7_status = 0;
	etage8_proc = 0.0;
	etage8_status = 0;
	etage9_proc = 0.0;
	etage9_status = 0;
	DefineAnimations ();
	// physical specs
	SetEmptyMass (1000000);
	SetCW (1, 1, 1, 1);
	SetSize (60);
	SetCrossSections (_V(200,200,200));
	SetPMI (_V(20,20,20));
	SetTouchdownPoints (_V(0, -37, 10), _V(-10, -37, -10), _V(10, -37, -10));
	SetRotDrag (_V(1,1,1));
	if (GetFlightModel() >= 1) {
		SetPitchMomentScale (1e-3);
		SetBankMomentScale (1e-3);
	}
	SetMeshVisibilityMode ((AddMesh ("Proton24/ProtonTower")), MESHVIS_ALWAYS);
	SetCameraOffset (_V(0,28,25));

	PROPELLANT_HANDLE hProp = CreatePropellantResource (FUELMASS, FUELMASS);

	SURFHANDLE smoke = oapiRegisterExhaustTexture ("Vega_Missions/proton_tower_smoke");
	SURFHANDLE flame = oapiRegisterExhaustTexture ("Vega_Missions/proton_tower_flame");

	PARTICLESTREAMSPEC contrail_smoke = {
		0, 10, 15, 150, 0.85, 2.2, 30, 1.8, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 2,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 2,
		smoke
	};

	PARTICLESTREAMSPEC contrail_flame = {
		0, 20, 25, 150, 0.4, 0.3, 30, 2.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 2,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1,
		flame
	};

	PARTICLESTREAMSPEC contrail_smoke2 = {
		0, 6, 8, 70, 0, 2, 15, 3, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 2,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 2,
		smoke
	};

	PARTICLESTREAMSPEC contrail_flame2 = {
		0, 6, 8, 70, 0, 0.5, 10, 5, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 2,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1,
		flame
	};




	th_steam1 = CreateThruster (_V(-23,-37,-6), _V(0.7071,-0.7071,0), MAXTHRUST, hProp, ISP);
	th_steam2 = CreateThruster (_V(20,-37,-6), _V(-0.7071,-0.7071,0), MAXTHRUST, hProp, ISP);

	th_steam3 = CreateThruster (_V(-22,-37,-6), _V(0.34,-0.94,0), MAXTHRUST, hProp, ISP);
	th_steam4 = CreateThruster (_V(19,-37,-6), _V(-0.34,-0.94,0), MAXTHRUST, hProp, ISP);

	th_steam5 = CreateThruster (_V(-3,-36,-6), _V(0,1,0), MAXTHRUST, hProp, ISP);

	AddExhaustStream (th_steam5, &contrail_smoke2);
	AddExhaustStream (th_steam5, &contrail_flame2);

	AddExhaustStream (th_steam3, &contrail_smoke);
	AddExhaustStream (th_steam4, &contrail_smoke);
	AddExhaustStream (th_steam1, &contrail_smoke);
	AddExhaustStream (th_steam2, &contrail_smoke);
	AddExhaustStream (th_steam1, &contrail_flame);
	AddExhaustStream (th_steam2, &contrail_flame);





	prt_attach = CreateAttachment (false, prtattpos, _V(0,-1,0), _V(0,0,1), "PT");

}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new protontower (hvessel, flightmodel);
}
// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (protontower*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void protontower::clbkPostCreation ()
{
}
void protontower::clbkSaveState (FILEHANDLE scn)
{
	char cbuf[256];

	// default vessel parameters
	VESSEL2::clbkSaveState (scn);

	// custom parameters
	sprintf (cbuf, "%d %0.4f", twr_status, twr_proc);
	oapiWriteScenario_string (scn, "TWR", cbuf);
	sprintf (cbuf, "%d %0.4f", mach_status, mach_proc);
	oapiWriteScenario_string (scn, "MCH", cbuf);
	sprintf (cbuf, "%d %0.4f", etage1_status, etage1_proc);
	oapiWriteScenario_string (scn, "ET1", cbuf);
	sprintf (cbuf, "%d %0.4f", etage2_status, etage2_proc);
	oapiWriteScenario_string (scn, "ET2", cbuf);
	sprintf (cbuf, "%d %0.4f", etage3_status, etage3_proc);
	oapiWriteScenario_string (scn, "ET3", cbuf);
	sprintf (cbuf, "%d %0.4f", etage4_status, etage4_proc);
	oapiWriteScenario_string (scn, "ET4", cbuf);
	sprintf (cbuf, "%d %0.4f", etage5_status, etage5_proc);
	oapiWriteScenario_string (scn, "ET5", cbuf);
	sprintf (cbuf, "%d %0.4f", etage6_status, etage6_proc);
	oapiWriteScenario_string (scn, "ET6", cbuf);
	sprintf (cbuf, "%d %0.4f", etage7_status, etage7_proc);
	oapiWriteScenario_string (scn, "ET7", cbuf);
	sprintf (cbuf, "%d %0.4f", etage8_status, etage8_proc);
	oapiWriteScenario_string (scn, "ET8", cbuf);
	sprintf (cbuf, "%d %0.4f", etage9_status, etage9_proc);
	oapiWriteScenario_string (scn, "ET9", cbuf);
}
void protontower::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "TWR", 3)) {
			sscanf (line+3, "%d%lf", &twr_status, &twr_proc);
		} else if (!strnicmp (line, "MCH", 3)) {
			sscanf (line+3, "%d%lf", &mach_status, &mach_proc);
		} else if (!strnicmp (line, "ET1", 3)) {
			sscanf (line+3, "%d%lf", &etage1_status, &etage1_proc);
		} else if (!strnicmp (line, "ET2", 3)) {
			sscanf (line+3, "%d%lf", &etage2_status, &etage2_proc);
		} else if (!strnicmp (line, "ET3", 3)) {
			sscanf (line+3, "%d%lf", &etage3_status, &etage3_proc);
		} else if (!strnicmp (line, "ET4", 3)) {
			sscanf (line+3, "%d%lf", &etage4_status, &etage4_proc);
		} else if (!strnicmp (line, "ET5", 3)) {
			sscanf (line+3, "%d%lf", &etage5_status, &etage5_proc);
		} else if (!strnicmp (line, "ET6", 3)) {
			sscanf (line+3, "%d%lf", &etage6_status, &etage6_proc);
		} else if (!strnicmp (line, "ET7", 3)) {
			sscanf (line+3, "%d%lf", &etage7_status, &etage7_proc);
		} else if (!strnicmp (line, "ET8", 3)) {
			sscanf (line+3, "%d%lf", &etage8_status, &etage8_proc);
		} else if (!strnicmp (line, "ET9", 3)) {
			sscanf (line+3, "%d%lf", &etage9_status, &etage9_proc);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	SetAnimation (anim_twr, twr_proc);
	SetAnimation (anim_mach, mach_proc);
	SetAnimation (anim_etage1, etage1_proc);
	SetAnimation (anim_etage2, etage2_proc);
	SetAnimation (anim_etage3, etage3_proc);
	SetAnimation (anim_etage4, etage4_proc);
	SetAnimation (anim_etage5, etage5_proc);
	SetAnimation (anim_etage6, etage6_proc);
	SetAnimation (anim_etage7, etage7_proc);
	SetAnimation (anim_etage8, etage8_proc);
	SetAnimation (anim_etage9, etage9_proc);
}


int protontower::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {
  if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate))
  {
	switch (key)
	{
	case OAPI_KEY_NUMPAD1:
	if (etage1_status == 0) etage1_status = 1;
	else if (etage1_status == 1) etage1_status = 3;
	else if (etage1_status == 2) etage1_status = 3;
	else if (etage1_status == 3) etage1_status = 1;
	return 1;

	case OAPI_KEY_NUMPAD2:
	if (etage2_status == 0) etage2_status = 1;
	else if (etage2_status == 1) etage2_status = 3;
	else if (etage2_status == 2) etage2_status = 3;
	else if (etage2_status == 3) etage2_status = 1;
	return 1;

	case OAPI_KEY_NUMPAD3:
	if (etage3_status == 0) etage3_status = 1;
	else if (etage3_status == 1) etage3_status = 3;
	else if (etage3_status == 2) etage3_status = 3;
	else if (etage3_status == 3) etage3_status = 1;
	return 1;

	case OAPI_KEY_NUMPAD4:
	if (etage4_status == 0) etage4_status = 1;
	else if (etage4_status == 1) etage4_status = 3;
	else if (etage4_status == 2) etage4_status = 3;
	else if (etage4_status == 3) etage4_status = 1;
	return 1;

	case OAPI_KEY_NUMPAD5:
	if (etage5_status == 0) etage5_status = 1;
	else if (etage5_status == 1) etage5_status = 3;
	else if (etage5_status == 2) etage5_status = 3;
	else if (etage5_status == 3) etage5_status = 1;
	return 1;

	case OAPI_KEY_NUMPAD6:
	if (etage6_status == 0) etage6_status = 1;
	else if (etage6_status == 1) etage6_status = 3;
	else if (etage6_status == 2) etage6_status = 3;
	else if (etage6_status == 3) etage6_status = 1;
	return 1;

	case OAPI_KEY_NUMPAD7:
	if (etage7_status == 0) etage7_status = 1;
	else if (etage7_status == 1) etage7_status = 3;
	else if (etage7_status == 2) etage7_status = 3;
	else if (etage7_status == 3) etage7_status = 1;
	return 1;

	case OAPI_KEY_NUMPAD8:
	if (etage8_status == 0) etage8_status = 1;
	else if (etage8_status == 1) etage8_status = 3;
	else if (etage8_status == 2) etage8_status = 3;
	else if (etage8_status == 3) etage8_status = 1;
	return 1;

	case OAPI_KEY_NUMPAD9:
	if (etage9_status == 0) etage9_status = 1;
	else if (etage9_status == 1) etage9_status = 3;
	else if (etage9_status == 2) etage9_status = 3;
	else if (etage9_status == 3) etage9_status = 1;
	return 1;




	}
return 0;
}

  else { // unmodified keys
    switch (key) {
	case OAPI_KEY_K: // move twr
		if (twr_status == 0)twr_status = 1;
		else if (twr_status == 1)twr_status = 2;
		else if (twr_status == 2)twr_status = 1;
		else if (twr_status == 3)twr_status = 4;
		else if (twr_status == 4)twr_status = 5;
		else if (twr_status == 5)twr_status = 4;
		return 1;
	case OAPI_KEY_G: // machoires
		if (mach_status == 0) mach_status = 1;
		else if (mach_status == 1) mach_status = 3;
		else if (mach_status == 2) mach_status = 3;
		else if (mach_status == 3) mach_status = 1;
		return 1;
	}
  }
  return 0;
}

void protontower::SetSmoke()
{
VECTOR3 gpos1, gtwr;
Local2Global (_V(0,-35,0), gtwr);
ENGINESTATUS es;


for (DWORD i = 0; i < oapiGetVesselCount(); i++)
	{
	OBJHANDLE hV = oapiGetVesselByIndex (i);
	if (hV == GetHandle()) continue;
	oapiGetGlobalPos (hV, &gpos1);
		double prdist = dist(gpos1, gtwr);
		if (prdist < 90)
		{
		oapiGetEngineStatus(hV, &es);
		if (es.main > 0)
			{
			double englevel = (90-prdist)*0.02;
			if (englevel > 1)englevel = 1;
			if (englevel < 0)englevel = 0;
			SetThrusterLevel(th_steam1, englevel);
			SetThrusterLevel(th_steam2, englevel);
			SetThrusterLevel(th_steam3, englevel);
			SetThrusterLevel(th_steam4, englevel);
			SetThrusterLevel(th_steam5, englevel);
			}
		else
			{
			SetThrusterLevel(th_steam1, 0);
			SetThrusterLevel(th_steam2, 0);
			SetThrusterLevel(th_steam3, 0);
			SetThrusterLevel(th_steam4, 0);
			SetThrusterLevel(th_steam5, 0);
			}
		return;
		}
		else
			{
			SetThrusterLevel(th_steam1, 0);
			SetThrusterLevel(th_steam2, 0);
			SetThrusterLevel(th_steam3, 0);
			SetThrusterLevel(th_steam4, 0);
			SetThrusterLevel(th_steam5, 0);
			}
	}


}

void protontower::clbkPreStep(double SimT, double SimDT, double MJD)
{

	stepdt = SimDT;
	SetSmoke();


ENGINESTATUS es;
{
		VECTOR3 gpos, grms, pos, dir, rot;
		Local2Global (prtattpos, grms);  // global position of grapple point
		
		for (DWORD i = 0; i < oapiGetVesselCount(); i++) 
		{
			OBJHANDLE hV = oapiGetVesselByIndex (i);
			if (hV == GetHandle()) continue; // we don't want to grapple ourselves ...
			oapiGetGlobalPos (hV, &gpos);
			if (dist (gpos, grms) < (25 + 10)) { // in range
				VESSEL *v = oapiGetVesselInterface (hV);
				DWORD nAttach = v->AttachmentCount (true);
				for (DWORD j = 0; j < nAttach; j++) 
				{
					ATTACHMENTHANDLE hAtt = v->GetAttachmentHandle (true, j);
					const char *id = v->GetAttachmentId (hAtt);
					if (strncmp (id, "PR", 2)) continue; // attachment point not compatible
					v->GetAttachmentParams (hAtt, pos, dir, rot);
					v->Local2Global (pos, gpos);
					if (dist (gpos, grms) < 10) 
					{
						
						oapiGetEngineStatus(hV, &es);
						if (hV == GetAttachmentStatus (prt_attach))
						{
						if (es.main > 0.8)DetachChild (prt_attach);
						}
						
						else if (es.main == 0)
						{
						if (v->GetAttachmentStatus(hAtt) == NULL) AttachChild (hV, prt_attach, hAtt);
						//AttachChild (hV, prt_attach, hAtt);
						}
					return;
					}
				}
			}
		}

	}




}

void protontower::clbkPostStep (double simt, double simdt, double mjd)
{
	/////////////////// Animate tower
	if (twr_status == 1)
		{
		double dt = simdt * TOWER_SPEED;
		twr_proc = twr_proc + dt;
		if (twr_proc > 1.0)
			{
			twr_proc = 1.0;
			twr_status = 3;
			}
		SetAnimation (anim_twr, twr_proc);
		}

	else if (twr_status == 4)
		{
		double dt = simdt * TOWER_SPEED;
		twr_proc = twr_proc - dt;
		if (twr_proc < 0)
			{
			twr_proc = 0.0;
			twr_status = 0;
			}
		SetAnimation (anim_twr, twr_proc);
		}

		///////////////// Animate machoires

	if (mach_status == 1)
		{
		double dm = simdt * MACHOIRES_SPEED;
		mach_proc = mach_proc + dm;
		if (mach_proc > 1.0)
			{
			mach_proc = 1.0;
			mach_status = 2;
			}
		SetAnimation (anim_mach, mach_proc);
		}

	else if (mach_status == 3)
		{
		double dm = simdt * MACHOIRES_SPEED;
		mach_proc = mach_proc - dm;
		if (mach_proc < 0.0)
			{
			mach_proc = 0.0;
			mach_status = 0;
			}
		SetAnimation (anim_mach, mach_proc);
		}

	////////////////////// etages

	if (etage1_status == 1)
		{
		double dm = simdt * MACHOIRES_SPEED;
		etage1_proc = etage1_proc + dm;
		if (etage1_proc > 1.0)
			{
			etage1_proc = 1.0;
			etage1_status = 2;
			}
		SetAnimation (anim_etage1, etage1_proc);
		}

	else if (etage1_status == 3)
		{
		double dm = simdt * MACHOIRES_SPEED;
		etage1_proc = etage1_proc - dm;
		if (etage1_proc < 0.0)
			{
			etage1_proc = 0.0;
			etage1_status = 0;
			}
		SetAnimation (anim_etage1, etage1_proc);
		}


	if (etage2_status == 1)
		{
		double dm = simdt * MACHOIRES_SPEED;
		etage2_proc = etage2_proc + dm;
		if (etage2_proc > 1.0)
			{
			etage2_proc = 1.0;
			etage2_status = 2;
			}
		SetAnimation (anim_etage2, etage2_proc);
		}

	else if (etage2_status == 3)
		{
		double dm = simdt * MACHOIRES_SPEED;
		etage2_proc = etage2_proc - dm;
		if (etage2_proc < 0.0)
			{
			etage2_proc = 0.0;
			etage2_status = 0;
			}
		SetAnimation (anim_etage2, etage2_proc);
		}

	if (etage3_status == 1)
		{
		double dm = simdt * MACHOIRES_SPEED;
		etage3_proc = etage3_proc + dm;
		if (etage3_proc > 1.0)
			{
			etage3_proc = 1.0;
			etage3_status = 2;
			}
		SetAnimation (anim_etage3, etage3_proc);
		}

	else if (etage3_status == 3)
		{
		double dm = simdt * MACHOIRES_SPEED;
		etage3_proc = etage3_proc - dm;
		if (etage3_proc < 0.0)
			{
			etage3_proc = 0.0;
			etage3_status = 0;
			}
		SetAnimation (anim_etage3, etage3_proc);
		}


	if (etage4_status == 1)
		{
		double dm = simdt * MACHOIRES_SPEED;
		etage4_proc = etage4_proc + dm;
		if (etage4_proc > 1.0)
			{
			etage4_proc = 1.0;
			etage4_status = 2;
			}
		SetAnimation (anim_etage4, etage4_proc);
		}

	else if (etage4_status == 3)
		{
		double dm = simdt * MACHOIRES_SPEED;
		etage4_proc = etage4_proc - dm;
		if (etage4_proc < 0.0)
			{
			etage4_proc = 0.0;
			etage4_status = 0;
			}
		SetAnimation (anim_etage4, etage4_proc);
		}


	if (etage5_status == 1)
		{
		double dm = simdt * MACHOIRES_SPEED;
		etage5_proc = etage5_proc + dm;
		if (etage5_proc > 1.0)
			{
			etage5_proc = 1.0;
			etage5_status = 2;
			}
		SetAnimation (anim_etage5, etage5_proc);
		}

	else if (etage5_status == 3)
		{
		double dm = simdt * MACHOIRES_SPEED;
		etage5_proc = etage5_proc - dm;
		if (etage5_proc < 0.0)
			{
			etage5_proc = 0.0;
			etage5_status = 0;
			}
		SetAnimation (anim_etage5, etage5_proc);
		}


	if (etage6_status == 1)
		{
		double dm = simdt * MACHOIRES_SPEED;
		etage6_proc = etage6_proc + dm;
		if (etage6_proc > 1.0)
			{
			etage6_proc = 1.0;
			etage6_status = 2;
			}
		SetAnimation (anim_etage6, etage6_proc);
		}

	else if (etage6_status == 3)
		{
		double dm = simdt * MACHOIRES_SPEED;
		etage6_proc = etage6_proc - dm;
		if (etage6_proc < 0.0)
			{
			etage6_proc = 0.0;
			etage6_status = 0;
			}
		SetAnimation (anim_etage6, etage6_proc);
		}


	if (etage7_status == 1)
		{
		double dm = simdt * MACHOIRES_SPEED;
		etage7_proc = etage7_proc + dm;
		if (etage7_proc > 1.0)
			{
			etage7_proc = 1.0;
			etage7_status = 2;
			}
		SetAnimation (anim_etage7, etage7_proc);
		}

	else if (etage7_status == 3)
		{
		double dm = simdt * MACHOIRES_SPEED;
		etage7_proc = etage7_proc - dm;
		if (etage7_proc < 0.0)
			{
			etage7_proc = 0.0;
			etage7_status = 0;
			}
		SetAnimation (anim_etage7, etage7_proc);
		}


	if (etage8_status == 1)
		{
		double dm = simdt * MACHOIRES_SPEED;
		etage8_proc = etage8_proc + dm;
		if (etage8_proc > 1.0)
			{
			etage8_proc = 1.0;
			etage8_status = 2;
			}
		SetAnimation (anim_etage8, etage8_proc);
		}

	else if (etage8_status == 3)
		{
		double dm = simdt * MACHOIRES_SPEED;
		etage8_proc = etage8_proc - dm;
		if (etage8_proc < 0.0)
			{
			etage8_proc = 0.0;
			etage8_status = 0;
			}
		SetAnimation (anim_etage8, etage8_proc);
		}


	if (etage9_status == 1)
		{
		double dm = simdt * MACHOIRES_SPEED;
		etage9_proc = etage9_proc + dm;
		if (etage9_proc > 1.0)
			{
			etage9_proc = 1.0;
			etage9_status = 2;
			}
		SetAnimation (anim_etage9, etage9_proc);
		}

	else if (etage9_status == 3)
		{
		double dm = simdt * MACHOIRES_SPEED;
		etage9_proc = etage9_proc - dm;
		if (etage9_proc < 0.0)
			{
			etage9_proc = 0.0;
			etage9_status = 0;
			}
		SetAnimation (anim_etage9, etage9_proc);
		}





}