// ==============================================================
//                 ORBITER MODULE: tgo_bus
//                  
//
// tgo_bus.cpp
// Control module for tgo_bus vessel class
//
// 
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include "OrbiterSoundSDK40.h"
#include <cstdio>
#include <cmath>
#include <string>
#include <stdio.h>
#include <math.h>

// ==============================================================
// Some vessel parameters
// ==============================================================

const double ISP = 2900;
const double ISP_MAIN = 3300;
const double AV_ISP = 3223.5789;
const double RCSTH = 22;
const double MAINTH = 424;
const double CHUTE_SPEED = 1;
const double ARRAY_SPEED = 0.05;
const double ARRAY_ROT_SPEED = 0.025;
const double ANT_DEPLOY_SPEED = 0.1;
const double ANT_ROT_SPEED = 0.025;
const double FUEL_MASS = 2400;
const double EMP_MASS = 1332;
const double SCHP_MASS = 600;

#define YELLOW RGB(255, 255, 80)
#define GREEN RGB(0, 225, 0)

const VECTOR3 schp_off = _V(0,0,2.225);



// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
void VLiftCoeff (VESSEL *v, double aoa, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 9;
	static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD, -2*RAD, 10*RAD,15*RAD,50*RAD,90*RAD,180*RAD};
	static const double CL[nabsc]  = {0,0,0,0,0,0,0,0,0};
	static const double CM[nabsc]  = {0,0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
	*cl = CL[i] + (CL[i+1]-CL[i]) * f;  // aoa-dependent lift coefficient
	*cm = 0.0;
	*cd = 0.95 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}

void HLiftCoeff (VESSEL *v, double beta, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 8;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.95 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}


class tgo_bus: public VESSEL3 {
public:
	
	tgo_bus (OBJHANDLE hVessel, int flightmodel);
	~tgo_bus ();

	void clbkSetClassCaps (FILEHANDLE cfg);
	void DefineAnimations (void);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	//void clbkPostStep (double simt, double simdt, double mjd);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	void JettisonSchp(void);
	void CalcMass(void);
	void DoKillRot(void);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	int clbkConsumeDirectKey(char *kstate);
	int MyID;
	THRUSTER_HANDLE th_rcs[12], th_group[5], th_main[1];
	PROPELLANT_HANDLE hProp;
	ANIMATIONCOMPONENT_HANDLE Section, Section2, Section3, Section4, Section5, Section6, Section7;
	AIRFOILHANDLE hHA, hVA;

private:
	UINT anim_ant_deploy,anim_ant_rota, anim_ant_rotb, anim_solp_deploy, anim_solp_rot;
	double ant_deploy_proc, ant_rota_proc, ant_rotb_proc, solp_deploy_proc, solp_rot_proc;
	int  ant_status, solp_status, schp_status, kilrot_sw;


};

tgo_bus::tgo_bus (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{

}

tgo_bus::~tgo_bus ()
{
	
}

// Define animation sequences for moving parts
// --------------------------------------------------------------
void tgo_bus::DefineAnimations (void)
{
	// 1. Solar arrays - rotation
	anim_solp_rot = CreateAnimation (0);
	static UINT ArraySpRotGrp1[1] = {0};

	static MGROUP_ROTATE SpRot1 (0, ArraySpRotGrp1, 0, _V(0,0,0), _V(-1,0,0), (float)(PI2));
	Section=AddAnimationComponent (anim_solp_rot, 0.0, 1, &SpRot1);


	
	// 2. Solar arrays - folding
	anim_solp_deploy = CreateAnimation (0);
	static UINT ArrayLFoldGrp1[1] = {3};
	static UINT ArrayLFoldGrp2[1] = {0};
	static UINT ArrayLFoldGrp3[1] = {4};

	static UINT ArrayRFoldGrp1[1] = {2};
	static UINT ArrayRFoldGrp2[1] = {1};
	static UINT ArrayRFoldGrp3[1] = {5};


	static MGROUP_ROTATE ArrayLFold1 (0, ArrayLFoldGrp1, 1, _V(-1.05,0,0), _V(0,1,0), (float)(-PI*0.5));
	Section2=AddAnimationComponent (anim_solp_deploy, 0.0,   1, &ArrayLFold1, Section);

	static MGROUP_ROTATE ArrayLFold2 (0, ArrayLFoldGrp2, 1, _V(-1.066,0,1.392), _V(0,1,0), (float)(PI));
	Section3=AddAnimationComponent (anim_solp_deploy, 0.0,   1, &ArrayLFold2, Section2);

	static MGROUP_ROTATE ArrayLFold3 (0, ArrayLFoldGrp3, 1, _V(-1.095,0,-2.096), _V(0,1,0), (float)(-PI));
	AddAnimationComponent (anim_solp_deploy, 0.0,   1, &ArrayLFold3, Section3);

	//// STBD
	static MGROUP_ROTATE ArrayRFold1 (0, ArrayRFoldGrp1, 1, _V(1.05,0,0), _V(0,1,0), (float)(PI*0.5));
	Section4=AddAnimationComponent (anim_solp_deploy, 0.0,   1, &ArrayRFold1, Section);

	static MGROUP_ROTATE ArrayRFold2 (0, ArrayRFoldGrp2, 1, _V(1.066,0,1.392), _V(0,1,0), (float)(-PI));
	Section5=AddAnimationComponent (anim_solp_deploy, 0.0,   1, &ArrayRFold2, Section4);

	static MGROUP_ROTATE ArrayRFold3 (0, ArrayRFoldGrp3, 1, _V(1.095,0,-2.096), _V(0,1,0), (float)(PI));
	AddAnimationComponent (anim_solp_deploy, 0.0,   1, &ArrayRFold3, Section5);
	

	// 3. ant deploy
	anim_ant_deploy = CreateAnimation (0);
	static UINT AntGrp1[1] = {13};

	static MGROUP_ROTATE Ant1 (0, AntGrp1, 0, _V(0,1.664,-1.952), _V(1,0,0), (float)(-108*RAD));
	Section6=AddAnimationComponent (anim_ant_deploy, 0.0, 1, &Ant1);

	// 4. ant rota
	anim_ant_rota = CreateAnimation (0.34);
	static UINT AntGrp2[1] = {13};

	static MGROUP_ROTATE Ant2 (0, AntGrp2, 0, _V(0,1.664,-1.952), _V(1,0,0), (float)(-180*RAD));
	Section7=AddAnimationComponent (anim_ant_rota, 0.0, 1, &Ant2, Section6);

	// 5. ant rotb
	anim_ant_rotb = CreateAnimation (0);
	static UINT AntGrp3[3] = {12,13,14};

	static MGROUP_ROTATE Ant3 (0, AntGrp3, 3, _V(0,1.664,-1.952), _V(0,-0.309,0.951), (float)(PI2));
	AddAnimationComponent (anim_ant_rotb, 0.0, 1, &Ant3, Section7);


}



// ==============================================================
// tgo_bus class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void tgo_bus::clbkSetClassCaps (FILEHANDLE cfg)
{
	ant_deploy_proc = 0.0;
	ant_rota_proc = 0.34;
	ant_rotb_proc = 0.0;
	solp_deploy_proc = 0.0;
	solp_rot_proc = 0.0;
	ant_status = 0;
	solp_status = 0;
	schp_status = 0;
	kilrot_sw = 0;
	
	DefineAnimations ();
	// physical specs
	SetCW (1, 1, 1, 1);
	SetRotDrag (_V(1,1,1));
	SetPMI (_V(2,2,1));
	SetCrossSections (_V(7,7,4));
	SetSize (3);
	SetEmptyMass (EMP_MASS+SCHP_MASS);
	SetMeshVisibilityMode ((AddMesh ("TGO/tgo_bus")), MESHVIS_ALWAYS);
	SetMeshVisibilityMode ((AddMesh ("TGO/tgo_bkshl_stowed", &schp_off)), MESHVIS_ALWAYS);
	
	SetCameraOffset (_V(0,1.5,-1.8));
	SetTouchdownPoints (_V(0.1,0.8,-1.8), _V(0.512,-0.853,-1.8), _V(-0.512,-0.853,-1.8));
	SetSurfaceFrictionCoeff (1,1);
	SetDockParams (_V(0,0,-2), _V(0,0,-1), _V(0,1,0));

	// propellant resources

	hProp = CreatePropellantResource (FUEL_MASS, FUEL_MASS);

	
	// ***************** thruster definitions *******************
	th_main[0] = CreateThruster (_V( 0,0,0), _V(0,0, 1), MAINTH, hProp, ISP_MAIN);
	
	SURFHANDLE mainex = oapiRegisterExhaustTexture ("Exhaust2");
	AddExhaust (th_main[0], 3, 0.2, _V( 0,0,-2.67), _V(0,0,-1), mainex);
	
	
	
	
	
	
	th_rcs[ 0] = CreateThruster (_V(1.11, 0.79, -2.329), _V(-0.7071,0.7071,0), RCSTH, hProp, ISP);
	th_rcs[ 1] = CreateThruster (_V(1.11, -0.79, -2.329), _V(-0.7071,-0.7071,0), RCSTH, hProp, ISP);
	th_rcs[ 2] = CreateThruster (_V(-1.11, 0.79, -2.329), _V(0.7071,0.7071,0), RCSTH, hProp, ISP);
	th_rcs[ 3] = CreateThruster (_V(-1.11, -0.79, -2.329), _V(0.7071,-0.7071,0), RCSTH, hProp, ISP);

	th_rcs[ 8] = CreateThruster (_V(1.116, 0.808, -2.242), _V(-0.7071,0.7071,0), RCSTH, hProp, ISP);
	th_rcs[ 9] = CreateThruster (_V(1.116, -0.808, -2.242), _V(-0.7071,-0.7071,0), RCSTH, hProp, ISP);
	th_rcs[10] = CreateThruster (_V(-1.116, 0.808, -2.242), _V(0.7071,0.7071,0), RCSTH, hProp, ISP);
	th_rcs[11] = CreateThruster (_V(-1.11, -0.808, -2.242), _V(0.7071,-0.7071,0), RCSTH, hProp, ISP);
	
	th_rcs[ 4] = CreateThruster (_V(0.95, 0.95, -2.482), _V(0,0,1), RCSTH, hProp, ISP);
	th_rcs[ 5] = CreateThruster (_V(0.95,-0.95, -2.482), _V(0,0,1), RCSTH, hProp, ISP);
	th_rcs[ 6] = CreateThruster (_V(-0.95, 0.95, -2.482), _V(0,0,1), RCSTH, hProp, ISP);
	th_rcs[ 7] = CreateThruster (_V(-0.95,-0.95, -2.482), _V(0,0,1), RCSTH, hProp, ISP);
	

	SURFHANDLE tex_rcs = oapiRegisterExhaustTexture ("exhaust_atrcs");
	for(int i=0;i<12;i++)
	{
	AddExhaust (th_rcs[i], 2, 0.1, (tex_rcs));
	}

	th_group[0] = th_rcs[ 1];
	th_group[1] = th_rcs[ 3];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[ 0];
	th_group[1] = th_rcs[ 2];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[ 8];
	th_group[1] = th_rcs[11];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);

	th_group[0] = th_rcs[ 9];
	th_group[1] = th_rcs[10];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[ 2];
	th_group[1] = th_rcs[ 3];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[ 0];
	th_group[1] = th_rcs[ 1];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[ 4];
	th_group[1] = th_rcs[ 5];
	th_group[2] = th_rcs[ 6];
	th_group[3] = th_rcs[ 7];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_FORWARD);


	th_group[0] = th_rcs[ 4];
	th_group[1] = th_rcs[ 5];
	th_group[2] = th_rcs[ 6];
	th_group[3] = th_rcs[ 7];
	th_group[4] = th_main[ 0];
	CreateThrusterGroup (th_group, 5, THGROUP_MAIN);



	///////////////////////////////////////////////////////
	/////AirFoils


	hVA = CreateAirfoil3 (LIFT_VERTICAL, _V(0,0,0.1), VLiftCoeff, 0, 3, 3, 1);
	hHA = CreateAirfoil3 (LIFT_HORIZONTAL, _V(0,0,0.1), HLiftCoeff, 0, 3, 3, 1);

	

}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new tgo_bus (hvessel, flightmodel);
}




// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (tgo_bus*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void tgo_bus::clbkPostCreation ()
{
	
	MyID=ConnectToOrbiterSoundDLL(GetHandle());
	ReplaceStockSound(MyID,"Sound\\Vessel\\attsustain.wav",	REPLACE_MAIN_THRUST);
	
	CalcMass();

	if(schp_status == 1) DelMesh(1);

	if (solp_status == 2)
	{
	EditAirfoil (hVA,0x09, _V(0,0,0.1), 0, 3, 6, 1);
	EditAirfoil (hHA,0x09,  _V(0,0,0.1), 0, 3, 6, 1);
	}
}
void tgo_bus::clbkSaveState (FILEHANDLE scn)
{

	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState (scn);

	// custom parameters

	sprintf (cbuf, "%d %0.4f %0.4f %0.4f", ant_status, ant_deploy_proc, ant_rota_proc, ant_rotb_proc);
	oapiWriteScenario_string (scn, "ANT", cbuf);

	sprintf (cbuf, "%d %0.4f %0.4f", solp_status, solp_deploy_proc, solp_rot_proc);
	oapiWriteScenario_string (scn, "SLP", cbuf);

	sprintf (cbuf, "%d", schp_status);
	oapiWriteScenario_string (scn, "SCH", cbuf);
	
}
void tgo_bus::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;

	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "ANT", 3)) {
            sscanf (line+3, "%d%lf%lf%lf", &ant_status, &ant_deploy_proc, &ant_rota_proc, &ant_rotb_proc);
		}else if (!strnicmp (line, "SLP", 3)) {
            sscanf (line+3, "%d%lf%lf", &solp_status, &solp_deploy_proc, &solp_rot_proc);
		}else if (!strnicmp (line, "SCH", 3)) {
			sscanf_s (line+3, "%d", &schp_status);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	SetAnimation (anim_ant_deploy, ant_deploy_proc);
	SetAnimation (anim_ant_rota, ant_rota_proc);
	SetAnimation (anim_ant_rotb, ant_rotb_proc);
	
	SetAnimation (anim_solp_deploy, solp_deploy_proc);
	SetAnimation (anim_solp_rot, solp_rot_proc);
	
}






int tgo_bus::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {

bool cbSelectTarget(void *id, char *str, void *data);

  if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) 
	{
	  switch (key) 
	  {
		case OAPI_KEY_NUMPAD5: 
		if  (kilrot_sw == 0) kilrot_sw = 1;
		else kilrot_sw = 0;
		return 1;
	  }
	  return true;
	} 
  else if (KEYMOD_CONTROL (kstate)) 
	{
  
	}
  else 
	{ // unmodified keys
    switch (key) 
		{
	case OAPI_KEY_G: 
		if (solp_status  == 0)solp_status = 1;
		else if (solp_status  == 2)solp_status = 3;
		else if (solp_status  == 3)solp_status = 2;
		return 1;

		case OAPI_KEY_K: 
		if (ant_status == 0)ant_status = 1;
		else if (ant_status == 1)ant_status = 3;
		else if (ant_status == 2)ant_status = 3;
		else if (ant_status == 3)ant_status = 1;
		return 1;

		case OAPI_KEY_J: 
		if (schp_status == 0)JettisonSchp();
		return 1;
		}
	}
  
 return 0;
 }


int tgo_bus::clbkConsumeDirectKey(char *kstate)
{
		VECTOR3 avel = _V(0,0,0);
		GetAngularVel(avel);
		double dt = oapiGetSimStep();
		double drot = RAD*2*dt;

if (KEYMOD_SHIFT (kstate))
{
if (KEYDOWN (kstate, OAPI_KEY_NUMPAD2))
		{
		if (avel.x < RAD*10)
			{
			avel.x = avel.x+drot;
			if (avel.x > RAD*10) avel.x = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD8))
		{
		if (avel.x > -RAD*10)
			{
			avel.x = avel.x-drot;
			if (avel.x < -RAD*10) avel.x = -RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD1))
		{
		if (avel.y < RAD*10)
			{
			avel.y = avel.y+drot;
			if (avel.y > RAD*10) avel.y = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD3))
		{
		if (avel.y > -RAD*10)
			{
			avel.y = avel.y-drot;
			if (avel.y < -RAD*10) avel.y = -RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD6))
		{
		if (avel.z < RAD*10)
			{			
			avel.z = avel.z+drot;
			if (avel.z > RAD*10) avel.z = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD4))
		{
		if (avel.z > -RAD*10)
			{
			avel.z = avel.z-drot;
			if (avel.z < -RAD*10) avel.z = -RAD*10;
			}		
		}

		SetAngularVel(avel);
	}
return 0;
}


void tgo_bus::DoKillRot()
{
		VECTOR3 avel = _V(0,0,0);
		GetAngularVel(avel);
		double dt = oapiGetSimStep();
		double drot = RAD*2*dt;

		if (drot > abs(avel.x)) avel.x = 0;
		else avel.x = avel.x + (drot * ((abs(avel.x))/-avel.x));
		if (drot > abs(avel.y)) avel.y = 0;
		else avel.y = avel.y + (drot * ((abs(avel.y))/-avel.y));
		if (drot > abs(avel.z)) avel.z = 0;
		else avel.z = avel.z + (drot * ((abs(avel.z))/-avel.z));

		SetAngularVel (avel);
		if (avel.x == 0 && avel.y == 0 && avel.z == 0) kilrot_sw = 0;
}


bool tgo_bus::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{

	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);


	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	//int w5 = (int)(264*d);
	//int w6 = (int)(280*d);
	//int w7 = (int)(296*d);
	//int w8 = (int)(312*d);
	//int w9 = (int)(328*d);
	//int w10 = (int)(344*d);
	//int w11 = (int)(360*d);
	//int w12 = (int)(376*d);
	//int w13 = (int)(392*d);
	//int w14 = (int)(408*d);
	//int w15 = (int)(424*d);

	{
	skp->SetTextColor(0x0066FF66);
	double m1 = GetEmptyMass();
	double m2 = GetPropellantMass(hProp);
	double mass_fraction = ((m1+m2)/(m1));
	double dv = (AV_ISP*(log(mass_fraction)));
	char cbuf[256];
	sprintf(cbuf,"Current Available dV %0.2f m/s", dv);
	skp->Text ((10+roxl), (wc+royl), cbuf,strlen(cbuf));
	}

	if(schp_status == 0)
	{
	double m1 = GetEmptyMass();
	double m2 = GetPropellantMass(hProp);
	double mass_fraction = ((m1+m2-SCHP_MASS)/(m1-SCHP_MASS));
	double dv = (AV_ISP*(log(mass_fraction)));
	char cbuf[256];
	sprintf(cbuf,"Post-Sep Available dV %0.2f m/s", dv);
	skp->Text ((10+roxl), (w0+royl), cbuf,strlen(cbuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char hbuf[256];
	sprintf(hbuf,"Shift+NumPad = Reaction Wheel Att.Control");
	skp->Text ((10+roxl), (w1+royl), hbuf,strlen(hbuf));
	}

	if (kilrot_sw == 1)
	{
	skp->SetTextColor(0x0066FFFF);
	char gbuf[256];
	sprintf(gbuf,"                                          Kill Rot.Active");
	skp->Text ((10+roxl), (w1+royl), gbuf,strlen(gbuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"G = Deploy/Lock Solar Panels");
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"K = Deploy/Stow HGA");
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}

	if (schp_status == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"J = Jettison EDL Capsule");
	skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
	}


	


	

	return true;
}


void tgo_bus::clbkPreStep(double SimT, double SimDT, double MJD) {
	//sprintf(oapiDebugString(), "ps fm %d", GetFlightModel());

	if (kilrot_sw == 1) DoKillRot();
	
	///////// animations

	// deploy solp
	if (solp_status == 0) solp_deploy_proc = 0;
	else if (solp_status == 1) 
	{
		double da = SimDT * ARRAY_SPEED;
		solp_deploy_proc = solp_deploy_proc + da;
		if (solp_deploy_proc > 1) 
		{
			solp_deploy_proc = 1;
			solp_status = 2;
			EditAirfoil (hVA,0x09, _V(0,0,0.1), 0, 3, 6, 1);
			EditAirfoil (hHA,0x09,  _V(0,0,0.1), 0, 3, 6, 1);
		}
	}
	else if (solp_status > 1)
	{
		solp_deploy_proc = 1;
	
	

	//////// suntracking
	VECTOR3 sun_pos_glob = _V(0,0,0);
	VECTOR3 sun_pos_loc = _V(0,0,0);

	Global2Local(sun_pos_glob, sun_pos_loc);
	double sun_angle = atan2(-sun_pos_loc.y, -sun_pos_loc.z);
	if(sun_angle < 0) sun_angle = sun_angle + PI2;

	if (solp_status == 3)sun_angle = PI;
	double sun_angle_proc = sun_angle/PI2;

	int rot_dir = 1;
	double proc_diff = (sun_angle_proc - solp_rot_proc);
	if (proc_diff > 0.5) rot_dir = -1;
	if (proc_diff < 0) rot_dir = -1;
	if (proc_diff < -0.5) rot_dir = 1;

	double da = SimDT * ARRAY_ROT_SPEED;
	if (proc_diff > 0.5)proc_diff = proc_diff - 1; // PI PI2
	else if (proc_diff < -0.5)proc_diff = proc_diff + 1;
	
	if (da > abs(proc_diff)) solp_rot_proc = sun_angle_proc;
	else solp_rot_proc = solp_rot_proc + (da*rot_dir);
	
	int srpi = (int) (solp_rot_proc);
	if (solp_rot_proc > 1) solp_rot_proc = solp_rot_proc - srpi;
	else if (solp_rot_proc < 0) solp_rot_proc = solp_rot_proc + 1 - srpi;
	SetAnimation (anim_solp_rot, solp_rot_proc);
	}
	SetAnimation (anim_solp_deploy, solp_deploy_proc);

	/////////////////////////////////////////////////////// ant
	if (ant_status == 0) ant_deploy_proc = 0;
	else if (ant_status == 1) 
	{
		double da = SimDT * ANT_DEPLOY_SPEED;
		ant_deploy_proc = ant_deploy_proc + da;
		if (ant_deploy_proc > 1) 
		{
			ant_deploy_proc = 1;
			ant_status = 2;
		}
		SetAnimation (anim_ant_deploy, ant_deploy_proc);
	}
	else if (ant_status == 2) 
	{
	ant_deploy_proc = 1;
	////////////////////////////////// earth tracking
	VECTOR3 earth_pos_glob = _V(0,0,0);
	VECTOR3 earth_pos_loc = _V(0,0,-1);
	OBJHANDLE hEarth = oapiGetObjectByName("Earth");
	if (hEarth)
	{
		oapiGetGlobalPos(hEarth, &earth_pos_glob);
		Global2Local(earth_pos_glob, earth_pos_loc);
	}
	normalise(earth_pos_loc);
	
	VECTOR3 req_shaft_axis = crossp(earth_pos_loc, _V(1,0,0));
	normalise(req_shaft_axis);
	VECTOR3 req_shaft_axis_normal = crossp(req_shaft_axis, _V(1,0,0));
	normalise(req_shaft_axis_normal);
	//sprintf(oapiDebugString(), "req_shaft_axis %0.4f %0.4f %0.4f", req_shaft_axis.x, req_shaft_axis.y, req_shaft_axis.z);


	double req_shaft_angle = atan2(-req_shaft_axis.z, req_shaft_axis.y);
	if (req_shaft_angle < 0) req_shaft_angle = req_shaft_angle+PI2;
	if (req_shaft_angle > (120*RAD)) req_shaft_angle = req_shaft_angle - PI;
	if (req_shaft_angle > (120*RAD)) req_shaft_angle = req_shaft_angle - PI;

	//
	double req_shaft_proc = 0.34 + (req_shaft_angle/PI);
	if (req_shaft_proc > 0.95)req_shaft_proc = 0.95;

	double dc = SimDT * ANT_DEPLOY_SPEED * 0.5;
	if (dc > (abs(ant_rota_proc - req_shaft_proc)))ant_rota_proc = req_shaft_proc;
	else if (req_shaft_proc > ant_rota_proc) ant_rota_proc = ant_rota_proc + dc;
	else if (req_shaft_proc < ant_rota_proc) ant_rota_proc = ant_rota_proc - dc;


	///////////////// 

	double da = SimDT * ANT_ROT_SPEED;
	
	//// create rotation matrix
	VECTOR3 rot_vec = _V(1, 0, 0);
	double rot_angle = PI*(req_shaft_proc-0.34);

	double ct = cos(rot_angle);
	double st = sin(rot_angle);

	double ux = rot_vec.x;
	double uy = rot_vec.y;
	double uz = rot_vec.z;
	double ux2 = ux*ux;
	double uy2 = uy*uy;
	double uz2 = uz*uz;

	MATRIX3 RG;

	RG.m11 = ux2+(1-ux2)*ct;
	RG.m12 = ux*uy*(1-ct)-uz*st;
	RG.m13 = ux*uz*(1-ct)+uy*st;
	RG.m21 = ux*uy*(1-ct)+uz*st;
	RG.m22 = uy2+(1-uy2)*ct;
	RG.m23 = uy*uz*(1-ct)-ux*st;
	RG.m31 = ux*uz*(1-ct)-uy*st;
	RG.m32 = uy*uz*(1-ct)+ux*st;
	RG.m33 = uz2+(1-uz2)*ct;

	VECTOR3 earth_pos_loc_rot = mul(RG, earth_pos_loc);

	double req_dish_angle = atan2(-earth_pos_loc_rot.x,-earth_pos_loc_rot.z);
	if (req_dish_angle < 0) req_dish_angle = req_dish_angle + PI2;
	double req_dish_proc = req_dish_angle/PI2;
	//sprintf(oapiDebugString(), "req_dish_angle %0.4f",req_dish_angle*DEG);

	int rotb_dir = 1;
	double proc_diff = (req_dish_proc - ant_rotb_proc);
	if (proc_diff > 0.5) rotb_dir = -1;
	if (proc_diff < 0) rotb_dir = -1;
	if (proc_diff < -0.5) rotb_dir = 1;

	if (da > abs(proc_diff))ant_rotb_proc = req_dish_proc;
	else ant_rotb_proc = ant_rotb_proc + (rotb_dir*da);

	int rtbi = (int) (ant_rotb_proc);
	if (ant_rotb_proc > 1) ant_rotb_proc = ant_rotb_proc - rtbi;
	else if (ant_rotb_proc < 0) ant_rotb_proc = ant_rotb_proc + 1 - rtbi;


	SetAnimation (anim_ant_deploy, ant_deploy_proc);
	SetAnimation (anim_ant_rota, ant_rota_proc);
	SetAnimation (anim_ant_rotb, ant_rotb_proc);
	}
	//////////////////////////////////////////
	else if (ant_status == 3) 
	{
	double da = SimDT * ANT_ROT_SPEED;
	int rot_dir = -1;

	if (ant_rotb_proc < 0.5) ant_rotb_proc = ant_rotb_proc-da;
	else ant_rotb_proc = ant_rotb_proc+da;

	if (ant_rotb_proc < 0) ant_rotb_proc = 0;
	else if (ant_rotb_proc > 1) ant_rotb_proc = 0;

	if (ant_rotb_proc == 0)
	{
	da = SimDT * ANT_DEPLOY_SPEED * 0.5;
	if (ant_rota_proc > 0.34) 
		{
		ant_rota_proc = ant_rota_proc-da;
		if (ant_rota_proc < 0.34)ant_rota_proc = 0.34;
		}
	else if (ant_rota_proc < 0.34) 
		{
		ant_rota_proc = ant_rota_proc+da;
		if (ant_rota_proc > 0.34)ant_rota_proc = 0.34;
		}
	}

	if (ant_rota_proc == 0.34 && ant_rotb_proc == 0)
	{
		da = SimDT * ANT_DEPLOY_SPEED;
		ant_deploy_proc = ant_deploy_proc - da;
		if (ant_deploy_proc < 0) 
		{
			ant_deploy_proc = 0;
			ant_status = 0;
		}
	}
	SetAnimation (anim_ant_deploy, ant_deploy_proc);
	SetAnimation (anim_ant_rota, ant_rota_proc);
	SetAnimation (anim_ant_rotb, ant_rotb_proc);
	}
}
//void tgo_bus::clbkPostStep (double simt, double simdt, double mjd)
//{

//}

////////// JETTISONS //////////////////////////////

void tgo_bus::JettisonSchp()
{
	VECTOR3 ofs = schp_off;
	VECTOR3 vofs = _V(0,0,0.37);

	char name[256];
	char classname[256];
	strcpy (classname, "TGO/tgo_edl_capsule");
	strcpy(name, "EDM_Capsule");

	
  VESSELSTATUS vs;
  GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (ofs, vs.rpos);
  GlobalRot (vofs, rvel);
  vs.rvel = vs.rvel + rvel;
  vs.vrot.z = vs.vrot.z + (15*RAD);
  vs.status = 0;
  oapiCreateVessel (name, classname, vs);

  
  schp_status = 1;
  CalcMass();
  DelMesh(1);
}

void tgo_bus::CalcMass()
{
	double current_mass = EMP_MASS;
	if(schp_status == 0)current_mass = current_mass + SCHP_MASS;
	SetEmptyMass(current_mass);
}

