// ==============================================================
//                 ORBITER MODULE: tgo_edl_capsule
//                  
//
// tgo_edl_capsule.cpp
// Control module for tgo_edl_capsule vessel class
//
// 
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
//#include "OrbiterSoundSDK40.h"
//#include <cstdio>
//#include <cmath>
//#include <string>
#include <stdio.h>
#include <math.h>

// ==============================================================
// Some vessel parameters
// ==============================================================


const double CHUTE_OPERATING_SPEED = 1;
const double COLLAPSE_SPEED = 0.5;
const double BKSHL_MASS = 240;
const double HTSHLD_MASS = 80;
const double SCHP_MASS = 280;
const double par_size = 24;
const double par_ref = 3;
const double cap_size = 3;
const double cap_ref = 0.1;
const VECTOR3 schp_off = _V(0,0,-0.1);

static TOUCHDOWNVTX tdvtx_geardown[3] = {
	{_V(0, 2,0.001), 1e4, 1e4, 0.5, 0.5},
	{_V(-1.732, -1, 0.001), 1e4, 1e4, 0.5, 0.5},
	{_V( 1.732, -1, 0.001), 1e4, 1e4, 0.5, 0.5}
};

//#define BOOM	1


// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
void VLiftCoeff (VESSEL *v, double aoa, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 9;
	static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD, -2*RAD, 10*RAD,15*RAD,50*RAD,90*RAD,180*RAD};
	static const double CL[nabsc]  = {0,0,0,0,0,0,0,0,0};
	static const double CM[nabsc]  = {0,0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
	*cl = CL[i] + (CL[i+1]-CL[i]) * f;  // aoa-dependent lift coefficient
	*cm = 0.0;
	*cd = 0.95 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}

void HLiftCoeff (VESSEL *v, double beta, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 8;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.95 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}


class tgo_edl_capsule: public VESSEL3 {
public:
	
	tgo_edl_capsule (OBJHANDLE hVessel, int flightmodel);
	~tgo_edl_capsule ();

	void clbkSetClassCaps (FILEHANDLE cfg);
	void DefineAnimations (void);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	void JettisonHtshld(void);
	void JettisonSchp(void);
	void CalcMass(void);
	void DoBoom(void);
	//int MyID;
	

private:
	UINT anim_chute, anim_col;
	double chute_proc, col_proc, boom_proc, pstream_lvl;
	int chute_status, col_status,htshld_status, schp_status, boom_status;
	AIRFOILHANDLE hHA, hVA;
	ATTACHMENTHANDLE hSchp_att;
	MESHHANDLE hChute;
	//PSTREAM_HANDLE pHndl;
	//int report_sw;
};

tgo_edl_capsule::tgo_edl_capsule (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
}

tgo_edl_capsule::~tgo_edl_capsule ()
{
	//if (tex_ion) oapiDestroySurface(tex_ion);
	//if (tex_rcs) oapiDestroySurface(tex_rcs);
}

// Define animation sequences for moving parts
// --------------------------------------------------------------
void tgo_edl_capsule::DefineAnimations (void)
{
	// 1. chute reefed
	static UINT HiGainAnt1Grp[2] = {0,1};
	static MGROUP_SCALE HiGainAnt1 (3, HiGainAnt1Grp, 2, _V(0,0,-0.2),_V(0.01,0.01,0.005));
	anim_chute = CreateAnimation (0);
	AddAnimationComponent (anim_chute, 0.9, 1, &HiGainAnt1);


	
	// 2. unreef
	static UINT HiGainAnt2Grp[2] = {0,1};
	static MGROUP_SCALE HiGainAnt2 (3, HiGainAnt2Grp, 2, _V(0,0,-0.2),_V(0.5,0.5,1));
	AddAnimationComponent (anim_chute, 0, 0.9, &HiGainAnt2);

	// 3. collaspe rot
	anim_col = CreateAnimation (0);
	static UINT HiGainAnt3Grp[2] = {0,1};
	static MGROUP_ROTATE HiGainAnt3 (3, HiGainAnt3Grp, 2, _V(0,0,-0.05),_V(1,0,0), 80*RAD);
	AddAnimationComponent (anim_col, 0, 0.8, &HiGainAnt3);

	static UINT HiGainAnt4Grp[2] = {0,1};
	static MGROUP_SCALE HiGainAnt4 (3, HiGainAnt4Grp, 2, _V(0,0,-0.1),_V(1,1,0.01));
	AddAnimationComponent (anim_col, 0.8, 1, &HiGainAnt4);


}



// ==============================================================
// tgo_edl_capsule class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void tgo_edl_capsule::clbkSetClassCaps (FILEHANDLE cfg)
{
	chute_proc = 1;
	chute_status = 2;
	col_proc = 0;
	col_status = 0;
	htshld_status = 0;
	schp_status = 0;
	boom_proc = 0;
	pstream_lvl = 0;
	boom_status = 0;
	//report_sw = 0;

	DefineAnimations ();
	// physical specs
	SetCW (4, 4, 2,2);
	SetRotDrag (_V(2,2,0.1));
	SetPMI (_V(4,4,2));
	SetCrossSections (_V(2,2,4));
	SetSize (2);
	CalcMass();
	//SetClipRadius(5);
	SetMeshVisibilityMode ((AddMesh ("TGO/tgo_bkshl")), MESHVIS_EXTERNAL);
	SetMeshVisibilityMode ((AddMesh ("TGO/tgo_htshld")), MESHVIS_EXTERNAL);
	SetMeshVisibilityMode ((AddMesh ("TGO/tgo_cap")), MESHVIS_EXTERNAL);
	SetMeshVisibilityMode ((AddMesh ("TGO/tgo_bkshl_chute")), MESHVIS_NEVER);
	
	SetCameraOffset (_V(0,0,0.2));
	//SetTouchdownPoints (_V(0,0.8,0.001), _V(-0.512,-0.853,0.001), _V(0.512,-0.853,0.001));
	SetTouchdownPoints (tdvtx_geardown, 3);
	SetSurfaceFrictionCoeff (1,1);
	

	///////////////////////////////////////////////////////
	/////AirFoils


	hVA = CreateAirfoil3 (LIFT_VERTICAL, _V(0,0,-cap_ref), VLiftCoeff, 0, cap_size, cap_size, 1);
	hHA = CreateAirfoil3 (LIFT_HORIZONTAL, _V(0,0,-cap_ref), HLiftCoeff, 0, cap_size, cap_size, 1);

	// reentry pstream
	/*PARTICLESTREAMSPEC rps = {
		0, 2.5, 40, 0, 0.001, 0.2, 4, 0.2, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1, PARTICLESTREAMSPEC::ATM_PLOG, 2e7, 1e16
	};
	pHndl = AddReentryStream (&rps);*/

	////////////////////// boom pstreams

	SURFHANDLE exh = oapiRegisterExhaustTexture ("contrail3");

	PARTICLESTREAMSPEC big_pstream = {
		0, 2, 5, 10, 0.7, 5, 20, 4.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 8,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1
		};

	PARTICLESTREAMSPEC small_pstream = {
		0, 0.25, 1, 25, 2, 2, 0, 4.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
		exh};

	AddParticleStream(&small_pstream, _V(0,0,0), _V(0,0,-1), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(0,0,-1), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(0,1,0), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(0,-1,0), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(1,0,0), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(-1,0,0), &pstream_lvl);

	AddParticleStream(&big_pstream, _V(0,0,0), _V(0,0,1), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(0,0,-1), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(0,1,0), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(0,-1,0), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(1,0,0), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(-1,0,0), &pstream_lvl);


	////////// attachment for schp
	hSchp_att = CreateAttachment(false, schp_off, _V(0,0,1),  _V(0,1,0), "BS", false);
}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new tgo_edl_capsule (hvessel, flightmodel);
}




// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (tgo_edl_capsule*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void tgo_edl_capsule::clbkPostCreation ()
{
	CalcMass();

	if (chute_status < 2) 
	{
		EditAirfoil (hVA,0x09, _V(0,0,-(cap_ref+(par_ref*(1-chute_proc)))), 0, 0, (cap_size+(par_size*(1-chute_proc))), 0);
		EditAirfoil (hHA,0x09, _V(0,0,-(cap_ref+(par_ref*(1-chute_proc)))), 0, 0, (cap_size+(par_size*(1-chute_proc))), 0);
		if (chute_status == 0) SetSize(15);
		SetPMI (_V(4,4,2));
		SetMeshVisibilityMode (2, MESHVIS_NEVER);
		SetMeshVisibilityMode (3, MESHVIS_EXTERNAL);
	}

	if (htshld_status == 1) SetMeshVisibilityMode (1, MESHVIS_NEVER);


	SetAnimation (anim_chute, chute_proc);
	
	///////////////////////////////////////////
	//MyID=ConnectToOrbiterSoundDLL(GetHandle());
	////ReplaceStockSound(MyID,"Sound\\Vessel\\ambient5.wav",	REPLACE_WHEELBRAKE);
	//RequestLoadVesselWave(MyID,BOOM,"Sound\\Vessel\\sonicbangfar1.wav",BOTHVIEW_FADED_MEDIUM);
	//ReplaceStockSound(MyID,"Sound\\Vessel\\ambient5.wav",	REPLACE_LAND_TOUCHDOWN);
	////ReplaceStockSound(MyID,"NULL",	REPLACE_LAND_TOUCHDOWN);
	////SoundOptionOnOff(MyID,PLAYMAINTHRUST,FALSE);
	////SoundOptionOnOff(MyID,PLAYRETROTHRUST,FALSE);
	////SoundOptionOnOff(MyID,PLAYHOVERTHRUST,FALSE);
	//SoundOptionOnOff(MyID,PLAYRADIOATC,FALSE);
	//SoundOptionOnOff(MyID,PLAYGPWS,FALSE);
	//SoundOptionOnOff(MyID,PLAYCOUNTDOWNWHENTAKEOFF,FALSE);
	////SoundOptionOnOff(MyID,PLAYLANDINGANDGROUNDSOUND,FALSE);
	//SoundOptionOnOff(MyID,PLAYCABINAIRCONDITIONING,FALSE);
	//SoundOptionOnOff(MyID,PLAYCABINRANDOMAMBIANCE,FALSE);
	//SoundOptionOnOff(MyID,DISPLAYTIMER,FALSE);
	//SoundOptionOnOff(MyID,PLAYDOCKLANDCLEARANCE,FALSE);
	//SoundOptionOnOff(MyID,PLAYRADARBIP,FALSE);

	if (schp_status == 3)
	{
		SetMeshVisibilityMode (2, MESHVIS_NEVER);
		SetMeshVisibilityMode (0, MESHVIS_NEVER);
	}
}
void tgo_edl_capsule::clbkSaveState (FILEHANDLE scn)
{

	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState (scn);

	// custom parameters
	
	sprintf (cbuf, "%d %d %d %0.4f %d %0.4f", htshld_status, schp_status, chute_status, chute_proc, col_status, col_proc);
	oapiWriteScenario_string (scn, "CHT", cbuf);
}
void tgo_edl_capsule::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;

	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "CHT", 3)) {
			sscanf (line+3, "%d%d%d%lf%d%lf", &htshld_status, &schp_status, &chute_status, &chute_proc, &col_status, &col_proc);
		} else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	SetAnimation (anim_chute, chute_proc);
	SetAnimation (anim_col, col_proc);	
	
}




void tgo_edl_capsule::clbkPreStep(double SimT, double SimDT, double MJD) {

	double alti = GetAltitude(ALTMODE_GROUND);
	double aspd = 0;
	aspd = GetAirspeed();
	////////////////////////// boom
	if (boom_status == 0)
	{
	if (GetDrag() > (220*600)) boom_status = 1;
	}
	if (boom_status == 1) DoBoom();
	
	/////////////////////////////// lander
	if (schp_status == 0)
	{
		OBJHANDLE hChild = GetAttachmentStatus(hSchp_att);
		if (hChild == NULL) JettisonSchp();
	}
	if (schp_status == 1)
	{
		if (alti < 1200 && aspd < 75) 
		{
		OBJHANDLE hChild = GetAttachmentStatus(hSchp_att);
		if(DetachChild(hSchp_att) == true)
			{
			oapiSetFocusObject(hChild);
			schp_status = 2;
			CalcMass();
			}
		}
	}
	//////////// heatshield 
	if (alti < 6850 && aspd < 120)
	{
		if (htshld_status == 0) JettisonHtshld();
	}

	//////////////// CHUTES /////////////////////////////////////
	if (alti < 20)
	{
		//DWORD lsw = GetFlightStatus();
		
		//if (lsw > 0)
		if(GroundContact() == true)
		{
		if (col_status == 0 && chute_status == 0) col_status = 1;
		SetMeshVisibilityMode (3, MESHVIS_EXTERNAL);
		}
	}

	//sprintf(oapiDebugString(), "%d", ibsw);
	
	if (chute_status == 2 && schp_status < 3)
	{
	chute_proc = 1;
		if (alti < 11200 && aspd < 500)
		{
		chute_status = 1;
		SetMeshVisibilityMode (3, MESHVIS_EXTERNAL);
		SetMeshVisibilityMode (2, MESHVIS_NEVER);
		SetPMI (_V(8,8,4));
		}
	}

	// Animate chute
	else if (chute_status == 1) 
	{
		double da = SimDT * CHUTE_OPERATING_SPEED;
		chute_proc = chute_proc - da;
		EditAirfoil (hVA,0x09, _V(0,0,-(cap_ref+(par_ref*(1-chute_proc)))), 0, 0, (cap_size+(par_size*(1-chute_proc))), 0);
		EditAirfoil (hHA,0x09, _V(0,0,-(cap_ref+(par_ref*(1-chute_proc)))), 0, 0, (cap_size+(par_size*(1-chute_proc))), 0);


		if (chute_proc < 0) 
		{
		chute_proc = 0;
		chute_status = 0;
		EditAirfoil (hVA,0x09, _V(0,0,-(cap_ref+(par_ref*(1-chute_proc)))), 0, 0, (cap_size+(par_size*(1-chute_proc))), 0);
		EditAirfoil (hHA,0x09, _V(0,0,-(cap_ref+(par_ref*(1-chute_proc)))), 0, 0, (cap_size+(par_size*(1-chute_proc))), 0);
		SetSize(15);
		SetRotDrag (_V(4,4,2));
		}
	}
	else if (chute_status == 0) chute_proc = 0;
	SetAnimation (anim_chute, chute_proc);

	///// collapse chute
	if (col_status == 1)
	{
	double dc = SimDT * COLLAPSE_SPEED;
	col_proc = col_proc + dc;
	if (col_proc > 1)
		{
		col_proc = 1;
		col_status = 2;
		}
	SetAnimation (anim_col, col_proc);
	}
	else if (col_status == 2)
	{
	col_proc = 1;
	SetAnimation (anim_col, col_proc);
	}
	////////////////////////////////////////////////////
}
void tgo_edl_capsule::clbkPostStep (double simt, double simdt, double mjd)
{
	////////////// report
	/*double alti = GetAltitude();

	if(report_sw == 0 && alti < 125000)
	{
	VECTOR3 haspdv = _V(0,0,0);
	GetHorizonAirspeedVector(haspdv);
	double xz = length(_V(haspdv.x,0,haspdv.z));
	double fpa = atan2(haspdv.y,xz);
	fpa = fpa*DEG;
	
	double lngi = 0;
	double lati = 0;
	double radi = 0;
	oapiGetEquPos(GetHandle(),&lngi,&lati,&radi);

	
	double tgt_lng = -6.1*RAD;
	double tgt_lat = -1.9*RAD;
	double antang = oapiOrthodome(lngi, lati, tgt_lng, tgt_lat);
	antang = antang*DEG;
	report_sw = 1;
	sprintf(oapiDebugString(), "fpa %0.2f antang %0.2f alti %0.2f",fpa, antang, alti*0.001);
	}*/
	
}

void tgo_edl_capsule::CalcMass()
{
	double current_mass = BKSHL_MASS;
	if(htshld_status == 0)current_mass = current_mass + HTSHLD_MASS;
	if(schp_status < 2)current_mass = current_mass + SCHP_MASS;
	if(schp_status == 3)current_mass = 200;
	SetEmptyMass(current_mass);
}

////////// JETTISONS //////////////////////////////

void tgo_edl_capsule::JettisonHtshld()
{
	VECTOR3 ofs = _V(0,0,0);
	VECTOR3 vofs = _V(0,0,0);
	char name[256];
	char classname[256];
	strcpy (classname, "TGO/tgo_htshld");
	strcpy (name, "EDM_Heatshield");

	
  VESSELSTATUS vs;
  GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (ofs, vs.rpos);
  GlobalRot (vofs, rvel);
  vs.rvel = vs.rvel + rvel;
  vs.status = 0;
  oapiCreateVessel (name, classname, vs);

  SetMeshVisibilityMode (1, MESHVIS_NEVER);
  htshld_status = 1;
  CalcMass();

}

void tgo_edl_capsule::JettisonSchp()
{
	VECTOR3 ofs = _V(0,0,0);
	VECTOR3 vofs = _V(0,0,0);
	char name[256];
	char classname[256];
	strcpy (classname, "TGO/tgo_schp");
	strcpy (name, "Schiaparelli");

	
  VESSELSTATUS vs;
  GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (ofs, vs.rpos);
  GlobalRot (vofs, rvel);
  vs.rvel = vs.rvel + rvel;
  vs.status = 0;

  OBJHANDLE hSchp;
  hSchp = oapiCreateVessel (name, classname, vs);
  VESSEL2* v =(VESSEL2*)oapiGetVesselInterface(hSchp);
  PROPELLANT_HANDLE ph = v->GetPropellantHandleByIndex (0);
  double maxfuel = v->GetPropellantMaxMass(ph);
  ATTACHMENTHANDLE hChild_att;
  v->SetPropellantMass(ph,maxfuel);
  hChild_att = v->GetAttachmentHandle(true, 0);
  AttachChild(hSchp,hSchp_att, hChild_att);

  schp_status = 1;
  CalcMass();

}

void tgo_edl_capsule::DoBoom()
{
pstream_lvl = 1;
boom_proc = boom_proc+oapiGetSimStep();
if (boom_proc > 0.5)
	{
	
	pstream_lvl = 0;
	boom_proc = 0;
	boom_status = 2;
	//DelExhaustStream(pHndl);
	}
if (htshld_status < 3)
	{
	htshld_status = 3;
	SetMeshVisibilityMode (2, MESHVIS_NEVER);
	SetMeshVisibilityMode (0, MESHVIS_NEVER);
	}
if (schp_status < 3)
	{
	schp_status = 3;
	OBJHANDLE hSchp = GetAttachmentStatus(hSchp_att);
	oapiDeleteVessel(hSchp);
	//PlayVesselWave(MyID, BOOM, 0, 255);
	}
CalcMass();
}