// ==============================================================
//                 ORBITER MODULE: tgo_schiaparelli
//                  
//
// tgo_schiaparelli.cpp
// Control module for tgo_schiaparelli vessel class
//
// 
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
//#include "OrbiterSoundSDK40.h"
//#include <cstdio>
//#include <cmath>
//#include <string>
#include <stdio.h>
#include <math.h>

// ==============================================================
// Some vessel parameters
// ==============================================================



const double SCHP_MASS = 235;
const double FUEL_MASS = 45;
const double ISP = 2000;
const double RCS = 1200;

static TOUCHDOWNVTX tdvtx_geardown[3] = {
	{_V(0,-0.1,1), 1e4, 1e4, 0.5, 0.5},
	{_V(-0.866, -0.1, -0.5), 1e4, 1e4, 0.5, 0.5},
	{_V( 0.866, -0.1, -0.5), 1e4, 1e4, 0.5, 0.5}
};


// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
void VLiftCoeff (VESSEL *v, double aoa, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 9;
	static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD, -2*RAD, 10*RAD,15*RAD,50*RAD,90*RAD,180*RAD};
	static const double CL[nabsc]  = {0,0,0,0,0,0,0,0,0};
	static const double CM[nabsc]  = {0,0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
	*cl = CL[i] + (CL[i+1]-CL[i]) * f;  // aoa-dependent lift coefficient
	*cm = 0.0;
	*cd = 0.1 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}

void HLiftCoeff (VESSEL *v, double beta, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 8;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.1 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}


class tgo_schiaparelli: public VESSEL3 {
public:
	
	tgo_schiaparelli (OBJHANDLE hVessel, int flightmodel);
	~tgo_schiaparelli ();

	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	VECTOR3 RotMat(VECTOR3 CURR_VECT,double rot_angle);
	void DoDescent (void);
	//int MyID;
	

private:
	AIRFOILHANDLE hHA, hVA;
	ATTACHMENTHANDLE hEDL_att;
	THRUSTER_HANDLE th_rcs[3], th_group[3];;
	PROPELLANT_HANDLE hProp;
	VECTOR3 rcs_pos[3], rcs_dir[3];
	int active, bam_sw, pls;
	double rv, timer, pls_tmr;
};

tgo_schiaparelli::tgo_schiaparelli (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
}

tgo_schiaparelli::~tgo_schiaparelli ()
{
	
}



// ==============================================================
// tgo_schiaparelli class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void tgo_schiaparelli::clbkSetClassCaps (FILEHANDLE cfg)
{
	active = 0;
	rv = 0;
	timer = 0;
	pls_tmr = 0;
	bam_sw = 0;
	pls = 0;
	// physical specs
	SetEmptyMass(SCHP_MASS);
	SetCW (4, 4, 2,2);
	SetRotDrag (_V(2,2,1));
	SetPMI (_V(0.5,1,0.5));
	SetCrossSections (_V(0.5,0.5,2.8));
	SetSize (2);
	SetMeshVisibilityMode ((AddMesh ("TGO/tgo_lander")), MESHVIS_EXTERNAL);
	SetCameraOffset (_V(0,0,0));
	//SetTouchdownPoints (_V(0,-0.1,1), _V(-0.866, -0.1, -0.5), _V(0.866, -0.1, -0.5));
	SetTouchdownPoints (tdvtx_geardown, 3);
	//SetSurfaceFrictionCoeff (1,1);

	// propellant resources

	hProp = CreatePropellantResource (FUEL_MASS, FUEL_MASS);

	// ***************** thruster definitions *******************
	
	th_rcs[0] = CreateThruster (_V( 0,0,1), _V(0,1,0), RCS, hProp, ISP);
	th_rcs[1] = CreateThruster (_V(-0.8660254,0,-0.5), _V(0,1,0), RCS, hProp, ISP);
	th_rcs[2] = CreateThruster (_V(0.8660254,0,-0.5), _V(0,1,0), RCS, hProp, ISP);

	
	SURFHANDLE mainex = oapiRegisterExhaustTexture ("Exhaust2");
	double lng = 1.3;
	double wdt = 0.08;
	//rcs_pos[0] = _V( 0,-0.1, 1.05);
	rcs_pos[0] = _V( 0,-0.1, 0.895);
	rcs_dir[0] = _V(0,-0.988,0.156);
	rv = 9;
	rcs_pos[1] = RotMat(rcs_pos[0], rv);
	rcs_dir[1] = RotMat(rcs_dir[0], rv);
	rv = -9;
	rcs_pos[2] = RotMat(rcs_pos[0], rv);
	rcs_dir[2] = RotMat(rcs_dir[0], rv);
	AddExhaust (th_rcs[0], lng, wdt, rcs_pos[0],rcs_dir[0], mainex);
	AddExhaust (th_rcs[0], lng, wdt, rcs_pos[1],rcs_dir[1], mainex);
	AddExhaust (th_rcs[0], lng, wdt, rcs_pos[2],rcs_dir[2], mainex);

	rv = 120;
	rcs_pos[0] = RotMat(rcs_pos[0], rv);
	rcs_dir[0] = RotMat(rcs_dir[0], rv);
	rcs_pos[1] = RotMat(rcs_pos[1], rv);
	rcs_dir[1] = RotMat(rcs_dir[1], rv);
	rcs_pos[2] = RotMat(rcs_pos[2], rv);
	rcs_dir[2] = RotMat(rcs_dir[2], rv);
	AddExhaust (th_rcs[1], lng, wdt, rcs_pos[0],rcs_dir[0], mainex);
	AddExhaust (th_rcs[1], lng, wdt, rcs_pos[1],rcs_dir[1], mainex);
	AddExhaust (th_rcs[1], lng, wdt, rcs_pos[2],rcs_dir[2], mainex);

	rv = 120;
	rcs_pos[0] = RotMat(rcs_pos[0], rv);
	rcs_dir[0] = RotMat(rcs_dir[0], rv);
	rcs_pos[1] = RotMat(rcs_pos[1], rv);
	rcs_dir[1] = RotMat(rcs_dir[1], rv);
	rcs_pos[2] = RotMat(rcs_pos[2], rv);
	rcs_dir[2] = RotMat(rcs_dir[2], rv);
	AddExhaust (th_rcs[2], lng, wdt, rcs_pos[0],rcs_dir[0], mainex);
	AddExhaust (th_rcs[2], lng, wdt, rcs_pos[1],rcs_dir[1], mainex);
	AddExhaust (th_rcs[2], lng, wdt, rcs_pos[2],rcs_dir[2], mainex);

	th_group[0] = th_rcs[0];
	//th_group[1] = th_rcs[3];
	CreateThrusterGroup (th_group, 1, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[2];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[2];
	CreateThrusterGroup (th_group, 1, THGROUP_ATT_BANKLEFT);

	th_group[0] = th_rcs[1];
	CreateThrusterGroup (th_group, 1, THGROUP_ATT_BANKRIGHT);
	
	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[1];
	th_group[2] = th_rcs[2];
	CreateThrusterGroup (th_group, 3, THGROUP_MAIN);

	///////////////////////////////////////////////////////
	/////AirFoils


	hVA = CreateAirfoil3 (LIFT_VERTICAL, _V(0,0,0), VLiftCoeff, 0, 0.05, 0.05, 1);
	hHA = CreateAirfoil3 (LIFT_HORIZONTAL, _V(0,0,0), HLiftCoeff, 0, 0.05, 0.05, 1);

	// reentry pstream
	/*PARTICLESTREAMSPEC rps = {
		0, 2, 20, 0, 0.001, 0.3, 4, 0.2, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1, PARTICLESTREAMSPEC::ATM_PLOG, 1e7, 2e16
	};
	AddReentryStream (&rps);*/

	////////// attachment for schp
	hEDL_att = CreateAttachment(true, _V(0,0.11,0), _V(0,1,0),  _V(0,0,1), "BS", false);
}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new tgo_schiaparelli (hvessel, flightmodel);
}




// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (tgo_schiaparelli*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void tgo_schiaparelli::clbkPostCreation ()
{
	//MyID=ConnectToOrbiterSoundDLL(GetHandle());
	////ReplaceStockSound(MyID,"Sound\\Vessel\\ambient5.wav",	REPLACE_WHEELBRAKE);
	//ReplaceStockSound(MyID,"Sound\\Vessel\\ambient5.wav",	REPLACE_LAND_TOUCHDOWN);
	////ReplaceStockSound(MyID,"NULL",	REPLACE_LAND_TOUCHDOWN);
	////SoundOptionOnOff(MyID,PLAYMAINTHRUST,FALSE);
	////SoundOptionOnOff(MyID,PLAYRETROTHRUST,FALSE);
	////SoundOptionOnOff(MyID,PLAYHOVERTHRUST,FALSE);
	//SoundOptionOnOff(MyID,PLAYRADIOATC,FALSE);
	//SoundOptionOnOff(MyID,PLAYGPWS,FALSE);
	//SoundOptionOnOff(MyID,PLAYCOUNTDOWNWHENTAKEOFF,FALSE);
	////SoundOptionOnOff(MyID,PLAYLANDINGANDGROUNDSOUND,FALSE);
	//SoundOptionOnOff(MyID,PLAYCABINAIRCONDITIONING,FALSE);
	//SoundOptionOnOff(MyID,PLAYCABINRANDOMAMBIANCE,FALSE);
	//SoundOptionOnOff(MyID,DISPLAYTIMER,FALSE);
	//SoundOptionOnOff(MyID,PLAYDOCKLANDCLEARANCE,FALSE);
	//SoundOptionOnOff(MyID,PLAYRADARBIP,FALSE);
	
}
void tgo_schiaparelli::clbkSaveState (FILEHANDLE scn)
{

	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState (scn);

	// custom parameters
	
	sprintf (cbuf, "%d %d %0.4f", active, bam_sw, timer);
	oapiWriteScenario_string (scn, "DSC", cbuf);
}
void tgo_schiaparelli::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;

	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "DSC", 3)) {
			sscanf (line+3, "%d%d%lf", &active, &bam_sw, &timer);
		} else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	
	
}




void tgo_schiaparelli::clbkPreStep(double SimT, double SimDT, double MJD) {

	double alti = GetAltitude(ALTMODE_GROUND);
	double aspd = 0;
	aspd = GetAirspeed();

	if (active == 0)
	{
		OBJHANDLE hParent = GetAttachmentStatus(hEDL_att);
		if (hParent == NULL)active = 1;
	}
	else if (active == 1) 
	{
	timer = timer+SimDT;
	if (timer > 1)
		{
		active = 2;
		timer = 0;
		DoDescent();
		}
	}
	else if (active == 2) 
	{		
		DoDescent();
		if (bam_sw < 2)
		{
		timer = timer+SimDT;
		}
		pls_tmr = pls_tmr+SimDT;
	}
	else if (active == 3) SetThrusterGroupLevel(THGROUP_MAIN, 0);

	if (pls_tmr > 0.05)
	{
		if (pls == 0) pls = 1;
		else pls = 0;
		pls_tmr = 0;
	}
}

void tgo_schiaparelli::clbkPostStep (double simt, double simdt, double mjd)
{


}


VECTOR3 tgo_schiaparelli::RotMat(VECTOR3 CURR_VECT,double rot_angle)
{
	
	double cosra = cos(RAD*rot_angle);
	double sinra = sin(RAD*rot_angle);

	VECTOR3 result_vector = _V(0,0,0);
	result_vector.y = CURR_VECT.y;
	result_vector.x = (CURR_VECT.x*cosra) - (CURR_VECT.z*sinra);
	result_vector.z = (CURR_VECT.x*sinra) + (CURR_VECT.z*cosra);
	return result_vector;

}

void tgo_schiaparelli::DoDescent (void)
{
	double mass = GetMass();
	double alt = GetAltitude(ALTMODE_GROUND);
	double altmean = GetAltitude(ALTMODE_MEANRAD);
	VECTOR3 aspdv;
	//GetShipAirspeedVector(aspdv);
	GetGroundspeedVector(FRAME_LOCAL, aspdv);
	VECTOR3 avel;
	GetAngularVel(avel);
	VECTOR3 weightv;
	double weight;
	GetWeightVector(weightv);
	weight = length(weightv);
	VECTOR3 haspdv;
	//GetHorizonAirspeedVector(haspdv);
	GetGroundspeedVector(FRAME_HORIZON, haspdv);
	double pitch = GetPitch();
	/*double lat, lng, rdi = 0;
	oapiGetEquPos(GetHandle(),&lng, &lat, &rdi);
	VECTOR3 windvct = oapiGetWindVector(GetGravityRef(),lng,lat,altmean, 0) ;
	VECTOR3 wndvctloc = _V(0,0,0);
	HorizonInvRot(windvct,wndvctloc);
	aspdv = aspdv + wndvctloc;*/

	if (bam_sw == 0)
	{
		if (pitch > (20*RAD))
		{
		bam_sw = 1;
		timer = 0;
		}
		if (timer > 0.5)
		{
		bam_sw = 1;
		timer = 0;
		}

	}
	else if (bam_sw == 1)
	{
		if (pitch > (40*RAD))
		{
		bam_sw = 2;
		timer = 0;
		}
		if (timer > 2)
		{
		bam_sw = 2;
		timer = 0;
		}
	}

	if (alt < 6)
	{
		haspdv.x = 0;
		haspdv.z = 0;
	}

	


	if (alt < 2.5)
	{
		if (haspdv.y > 0)
		{
		SetThrusterGroupLevel(THGROUP_MAIN, 0);
		active = 3;
		pls = 1;
		DeactivateNavmode(NAVMODE_HLEVEL);
		//sprintf(oapiDebugString(), "active3 vs %0.4f alt %0.4f", haspdv.y, alt);
		return;
		}
		
	}

	if (alt < 6) ActivateNavmode(NAVMODE_HLEVEL);
	if (alt < 1.9)DeactivateNavmode(NAVMODE_HLEVEL);

	double torquez = 1200*0.8660254;
	double torquex = 1200*1.000;

	
	double pf = (1/(cos(pitch)));

	double bank = GetBank();
	double bf = (1/(cos(bank)));

	double inrtx = (0.5*mass); // mpx SetPMI (_V(0.3,0.3,0.2));
	double inrtz = (0.5*mass);

	double reqpx = 1*(-atan(aspdv.z/aspdv.y));
	double reqpz = 1*(atan(aspdv.x/aspdv.y));


	double timex = (2*reqpx/ avel.x);
	double accx = -avel.x/timex;

	double timez = (2*reqpz/ avel.z);
	double accz = -avel.z/timez;

	double maxaccx = torquex/inrtx;
	double maxaccz = torquez/inrtz;


	double a;
	double b;
	double f = 20;

	

	if (avel.x < 0) avel.x = -avel.x;

	if (timex < 0) a = (reqpx/f*PI);

	else if (timex > avel.x*10/maxaccx) a = (reqpx/f*PI);

	else a = accx/maxaccx;

	// do z axis

	if (avel.z < 0) avel.z = -avel.z;

	if (timez < 0) b = (reqpz/f*PI);

	else if (timez > avel.z*10/maxaccz) b = (reqpz/f*PI);

	else  b = accz/maxaccz;


	//
	double vofs = 1;
	double altofs = 4;

	if (alt < 5)
	{
	vofs = 0;
	altofs = 1.9;
	}

	double time = (2*(alt-altofs)/(haspdv.y+vofs));
	double acc = ((haspdv.y)/time);
	//if (alt < 2.1) acc = -haspdv.y/0.025;
	double force = ((mass*acc)+ weight);
	double englvl = min(1, bf*pf*force/3600);
	a = a/2;
	b = b/2;

	if (alt < 3.2) //2.2
	{
	a = 0;
	b = 0;
	}

	double lvl1 = max(0,(englvl + a + a/3));
	double lvl2 =  max(0,(englvl - a + b + a/3));
	double lvl3 =  max(0,(englvl - a - b + a/3));

	lvl1 = min(1,lvl1);
	lvl2 =  min(1,lvl2);
	lvl3 =  min(1,lvl3);
	//// pls adjust
	{
		lvl1 = 2*lvl1;
		lvl2 = 2*lvl2;
		lvl3 = 2*lvl3;
		
		double dl = 0;
		if(lvl1 > 1)
		{
		dl = lvl1-1;
		lvl1 = lvl1-dl;
		lvl2 = lvl2-dl;
		lvl3 = lvl3-dl;;
		}

		if(lvl2 > 1)
		{
		dl = lvl2 - 1;
		lvl1 = lvl1-dl;
		lvl2 = lvl2-dl;
		lvl3 = lvl3-dl;;
		}

		if(lvl3 > 1)
		{
		dl = lvl3 - 1;
		lvl1 = lvl1-dl;
		lvl2 = lvl2-dl;
		lvl3 = lvl3-dl;
		}
	}
	//////////////
	SetThrusterLevel(th_rcs[0], lvl1);
	SetThrusterLevel(th_rcs[1], lvl2);
	SetThrusterLevel(th_rcs[2], lvl3);
	

	

	if (bam_sw == 0)
	{
	SetThrusterLevel(th_rcs[0], 1);
	SetThrusterLevel(th_rcs[1], 0.8);
	SetThrusterLevel(th_rcs[2], 0.8);
	}
	else if (bam_sw == 1)
	{
	SetThrusterLevel(th_rcs[0], 1);
	SetThrusterLevel(th_rcs[1], 1);
	SetThrusterLevel(th_rcs[2], 1);
	}

	
	if (pls == 1)
	{
		SetThrusterLevel(th_rcs[0], 0);
		SetThrusterLevel(th_rcs[1], 0);
		SetThrusterLevel(th_rcs[2], 0);
	}

	//if(GetThrusterGroupLevel(THGROUP_MAIN) > 0)sprintf(oapiDebugString(), "vs %0.4f alt %0.4f", haspdv.y, alt);
	
}